"""Slices CLI authentication."""

import os
from typing import Annotated

import typer
from requests.exceptions import HTTPError

from . import console, err_console
from .config import auth_file_location, delete_auth_token, parse_auth_token, save_auth_token, set_config_string
from .context import Context
from .oauth2.flow import AuthenticationFlow, DeviceFlow

auth_app = typer.Typer(help="User authentication", no_args_is_help=True)


@auth_app.command(help="Authenticate and retrieve credentials")
def login(
    ctx: Context,
    open_browser: Annotated[bool, typer.Option(help="Automatically open browser to perform login.")] = False,
):
    """Authenticate and retrieve credentials."""

    if os.environ.get("SLICES_AUTH_TOKEN_PROVIDER"):
        err_console.print("[bold red]Authentication is handled by external token provider.")

    if not open_browser:
        flow = DeviceFlow()
        token = flow.perform_login_flow()
    else:
        flow = AuthenticationFlow()
        token = flow.perform_login_flow()

    if token:
        save_auth_token(token)

        auth_info = parse_auth_token(token)
        console.print(f"[bold green]Successfully logged in as '{auth_info['preferred_username']}' ")
        console.print(f":floppy_disk: Saved to {auth_file_location}")

        if ctx.obj.get_current_project_id() is None:
            console.print(
                ":backhand_index_pointing_right: "
                "Now select the project that you want to use with [blue]slices project use <project name>[/]"
            )

    else:
        err_console.print("[red]Logging in failed. :x:")
        raise typer.Exit(1)


@auth_app.command(help="Show user credentials.")
def show(
    ctx: Context,
):
    """Show user credentials."""
    user_info = ctx.obj.user_info

    if user_info:
        console.print(f"Logged in as '{user_info['preferred_username']}'")
    else:
        err_console.print("[red]Not logged in. :x:")
        raise typer.Exit(1)


@auth_app.command(help="Remove user credentials")
def logout():
    """Remove user credentials."""
    if os.environ.get("SLICES_AUTH_TOKEN_PROVIDER"):
        err_console.print("[red]Cannot logout when using an external token provider. :x:")
        raise typer.Exit(1)

    delete_auth_token()
    set_config_string("default", None, "project")
    console.print("[bold green]Successfully logged out.[/] :white_check_mark:")
    console.print(f":floppy_disk: Deleted {auth_file_location}")


@auth_app.command(help="Get credentials for a specific audience.", hidden=True)
def get_for_audience(
    ctx: Context,
    audience: str,
    with_project: Annotated[bool, typer.Option(help="Include project ID in the token.")] = True,
):
    """Print user credentials for a specific audience."""
    user_info = ctx.obj.user_info

    if not user_info:
        err_console.print("[red]Not logged in. :x:")
        raise typer.Exit(1)
    try:
        print(
            ctx.obj.api.get_access_token_for_audience(audience, ctx.obj.require_project_id() if with_project else None)
        )
    except HTTPError as e:
        try:
            error_response = e.response.json()
            err_console.print(
                f"[bold red]Error: Could not get token. :x:[/]\n[red]{error_response['error_description']}"
            )
        except:  # noqa: E722
            err_console.print("[bold red]Error: Could not get token. :x:")
            err_console.print(f"[red]{str(e.response.text)}")
        raise typer.Exit(1) from None


@auth_app.command(help="Get id_token for a specific audience.", hidden=True)
def id_token(
    ctx: Context,
    audience: str,
):
    """Print id_token for a specific audience."""

    print(ctx.obj.api.get_id_token_for_audience(audience, ctx.obj.require_project_id()))
