"""BI CLI."""

import json
import os
import socket
import webbrowser
from datetime import datetime, timedelta, timezone
from enum import Enum
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Optional

import typer
import yaml
from pydantic import (
    BaseModel,
    Field,
    FilePath,
    PlainSerializer,
    PlainValidator,
    ValidationError,
    ValidationInfo,
    field_validator,
    model_validator,
)
from pylev import levenshtein
from rich.prompt import Prompt
from rich.table import Table
from slices_bi_client.exceptions import ApiException, ForbiddenException, NotFoundException
from slices_bi_client.models.bare_metal_link_create import BareMetalLinkCreate
from slices_bi_client.models.compute_resource import ComputeResource
from slices_bi_client.models.compute_resource_create import ComputeResourceCreate
from slices_bi_client.models.network_interface_create import NetworkInterfaceCreate
from slices_bi_client.models.new_resources_request import NewResourcesRequest
from slices_bi_client.models.reboot_action import RebootAction
from slices_bi_client.models.resource_action import ResourceAction
from slices_bi_client.models.resource_state import ResourceState
from slices_bi_client.models.ticket import Ticket
from slices_bi_client.models.update_resources_request import UpdateResourcesRequest
from slices_bi_client.models.vmlab_backend_schemas_bi_resource_reboot_action_action import (
    VmlabBackendSchemasBiResourceRebootActionAction,
)
from slices_expauth_client.models.experiment import Experiment
from slices_expauth_client.models.experiment_update import ExperimentUpdate

from . import console, err_console
from .bi_context import BIContext, BIContextObj, _is_type_id
from .bi_diskimage import disk_images_app
from .bi_flavor import flavors_app
from .context import Context
from .progress import follow_one_progress, follow_progress
from .util.pagination import get_next_page_marker
from .util.ssh_known_hosts import SSHKnownHostsManager
from .utils import get_error_detail, is_type_id

if TYPE_CHECKING:
    from slices_bi_client.models.console_login_details import ConsoleLoginDetails
    from slices_bi_client.models.disk_image import DiskImage

bi_app = typer.Typer(short_help="Basic Infrastructure", no_args_is_help=True)
bi_app.add_typer(flavors_app, name="flavor", short_help="Flavors.")
bi_app.add_typer(disk_images_app, name="diskimage", short_help="Disk images.")


def bi_site_callback(ctx: Context, _param: typer.CallbackParam, value: str):
    """Validate bi_site value."""
    if value not in ctx.obj.api.config.bi:
        raise typer.BadParameter("Invalid BI Site ID")

    return value


@bi_app.callback(invoke_without_command=True)
def initialize_bi(
    ctx: Context,
    site_id: Annotated[str, typer.Option(envvar="SLICES_BI_SITE_ID", callback=bi_site_callback)] = "vmlab",
    cluster_id: Annotated[str, typer.Option(envvar="SLICES_BI_CLUSTER_ID")] = "default",
):
    """Initialize BI Context."""
    if ctx.obj.user_info is None:
        err_console.print("[red]You must be logged in. :x:")
        raise typer.Exit(code=1)

    ctx.obj = BIContextObj(
        api=ctx.obj.api,
        user_info=ctx.obj.user_info,
        site_id=site_id,
        cluster_id=cluster_id,
    )


def _get_experiment(ctx: BIContext, experiment_id) -> Experiment:
    if _is_type_id(experiment_id, required_type="exp"):
        exp_obj = ctx.obj.api.experiments_api.get_experiment(experiment_id)
    else:
        exp_obj = ctx.obj.api.get_experiment(experiment_id, project_id=ctx.obj.require_project_id())

    if not exp_obj:
        err_console.print(f"Could not retrieve experiment {experiment_id} :x:")
        raise typer.Exit(code=1)
    return exp_obj


def _get_registered_pubkeys(ctx: BIContext, experiment: Experiment) -> list[str]:
    """Get public keys of all members in the experiment."""

    experiment_members = ctx.obj.api.experiments_api.get_experiment_members(experiment.id)

    return [
        pubkey.public_key
        for member in experiment_members.experiment_members
        for pubkey in ctx.obj.api.pubkey_api.get_user_pubkeys(member.user_id).public_keys
    ]


@cache
def _find_flavor_id(
    ctx: BIContext,
    flavor_any_id: str,
    site_id: str | None = None,
) -> str:
    flavors_api = (
        ctx.obj.api.get_bi_flavors_api(site_id, ctx.obj.require_project_id()) if site_id else ctx.obj.bi_flavors_api
    )

    flavors_page = flavors_api.get_flavors()
    for flavor_inner in flavors_page.flavors:
        flavor = flavor_inner.actual_instance
        if flavor and (flavor_any_id == flavor.id or flavor_any_id.lower() == flavor.friendly_name.lower()):
            return flavor.id
    err_console.print(f"[red]Unknown flavor {flavor_any_id!r} :mag:")
    err_console.print("Check: [blue]slices bi flavor list")
    raise typer.Exit(code=1)


MAX_DISK_IMAGE_STRING_DISTANCE = 5


@cache
def _find_disk_image_id(
    ctx: BIContext,
    disk_image_any_id: str,
    site_id: str | None = None,
) -> str:
    closest_match: DiskImage | None = None
    closest_match_distance = -1
    closest_match_collision = False

    disk_image_api = (
        ctx.obj.api.get_bi_images_api(site_id, ctx.obj.require_project_id()) if site_id else ctx.obj.bi_image_api
    )

    disk_image_page = disk_image_api.get_disk_images()
    for disk_image in disk_image_page.disk_images:
        if disk_image.id == disk_image_any_id:
            return disk_image.id

        distance = levenshtein(disk_image.friendly_name, disk_image_any_id)
        if not closest_match or distance < closest_match_distance:
            closest_match = disk_image
            closest_match_distance = distance
            closest_match_collision = False
        elif distance == closest_match_distance:
            closest_match_collision = True

    if closest_match:
        if not closest_match_collision or closest_match_distance < MAX_DISK_IMAGE_STRING_DISTANCE:
            return closest_match.id

        err_console.print(f"[red]Unknown disk image {disk_image_any_id!r} :mag:")
        err_console.print(f"Did you mean: '{closest_match.friendly_name}' ?")
        raise typer.Exit(code=2)

    err_console.print(f"[red]Unknown disk image {disk_image_any_id!r} :mag:")
    err_console.print("Check: [blue]slices bi diskimage list")
    raise typer.Exit(code=1)


# cfr. https://docs.cloud-init.io/en/24.1/explanation/format.html#headers-and-content-types
CLOUD_INIT_PREFIXES = (
    "#cloud-config",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#cloud-config-data
    "#!",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#user-data-script
    "#cloud-boothook",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#cloud-boothook
    "#include",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#include-file
    "## template: jinja",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#jinja-template
    "Content-Type: multipart/mixed",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#mime-multi-part-archive
    "#cloud-config-archive",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#cloud-config-archive
    "#part-handler",  # https://docs.cloud-init.io/en/24.1/explanation/format.html#part-handler
    # the following option is not mentioned in the former table, but is mentioned in the vendor data:
    "#cloud-config-jsonp",  #  https://docs.cloud-init.io/en/24.1/explanation/vendordata.html
)


def _create_compute_resource_create(
    name: str,
    disk_image_id: str,
    flavor_id: str,
    userdata: str | None,
    public_ipv4: bool,
):
    return ComputeResourceCreate(
        friendly_name=name,
        disk_image_id=disk_image_id,
        flavor_id=flavor_id,
        userdata=userdata,
        network_interfaces=[
            NetworkInterfaceCreate(
                port_id="eth0",
                network_id="private" if not public_ipv4 else "public",
                dhcp4=None,
                dhcp6=None,
                addresses=None,
                nameservers=None,
                routes=None,
            )
        ],
    )


def _read_userdata(user_data: typer.FileText | None) -> str | None:
    if user_data:
        userdata = user_data.read()
        if not userdata.startswith(CLOUD_INIT_PREFIXES):
            assert user_data
            err_console.print(
                f"[orange]WARNING: Userdata file {user_data.name} does not start with a valid cloud-init prefix. :x:\n"
                f"Valid prefixes are: {', '.join(CLOUD_INIT_PREFIXES)}"
            )
        return userdata
    return None


@bi_app.command()
def create(
    ctx: BIContext,
    name: Annotated[str, typer.Option(..., help="Name of the resource.", prompt=True)],
    image: Annotated[str, typer.Option(..., help="ID of the VM image.", prompt=True)],
    experiment: Annotated[str, typer.Option(..., help="Experiment name or ID.", prompt=True)],
    flavor: Annotated[str, typer.Option(..., help="ID of the flavor")] = "m1.small",
    user_data: Annotated[Optional[typer.FileText], typer.Option(help="File with cloudinit user-data.")] = None,
    duration: Annotated[
        int, typer.Option(min=1, max=90 * 24, help="Lifetime of the resource in hours. Can be extended later on.")
    ] = 24,
    wait: Annotated[bool, typer.Option(help="Wait for resource to be created.")] = False,
    public_ipv4: Annotated[bool, typer.Option(help="Request a public IPv4.")] = False,
    count: Annotated[int, typer.Option(help="Number of resources to create.", show_default=False)] = 1,
):
    """Request compute resource."""

    expires_at = datetime.now(timezone.utc).replace(second=0, microsecond=0) + timedelta(hours=duration)

    project_id = ctx.obj.require_project_id()

    if is_type_id(experiment, type="exp"):
        exp = ctx.obj.api.get_experiment_by_id(experiment_id=experiment)
    else:
        exp = ctx.obj.api.get_or_create_experiment(
            experiment_name=experiment,
            project_id=project_id,
            expires_at=expires_at,
        )

    # if experiment expiration is shorter than requested duration, extend the experiment
    if expires_at > exp.expires_at:
        ctx.obj.api.experiments_api.update_experiment(exp.id, ExperimentUpdate(expires_at=expires_at))

    ssh_authorized_keys = _get_registered_pubkeys(ctx, exp)
    if not ssh_authorized_keys:
        err_console.print("[red]You need to have at least one public SSH key registered. :x:")
        err_console.print("Use: [blue]slices pubkey register")
        raise typer.Exit(code=1)

    disk_image_id = _find_disk_image_id(ctx, image)
    flavor_id = _find_flavor_id(ctx, flavor)

    resources: list[ComputeResourceCreate] = []
    userdata = _read_userdata(user_data)
    if count == 1:
        resources.append(_create_compute_resource_create(name, disk_image_id, flavor_id, userdata, public_ipv4))
    else:
        resources.extend(
            _create_compute_resource_create(f"{name}{i+1}", disk_image_id, flavor_id, userdata, public_ipv4)
            for i in range(count)
        )

    nrr = NewResourcesRequest(
        expires_at=expires_at,
        resources=resources,
        ssh_authorized_keys=ssh_authorized_keys,
        tickets=[
            Ticket(
                jwt=ctx.obj.api.get_id_token_for_audience(ctx.obj.api.config.bi[ctx.obj.site_id].audience, project_id)
            )
        ],
    )

    try:
        response = ctx.obj.get_bi_resources_api_by_experiment(exp).request_experiment_resources(
            project=project_id,
            experiment_id=exp.id,
            new_resources_request=nrr,
        )
    except ForbiddenException as err:
        err_console.print("[bold red]No permission to create resource[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None
    except ApiException as err:
        err_console.print("[bold red]Could not create resource[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None
    if not len(response.resources):
        err_console.print("[bold red]No resources were created[/] :x:")

    if wait:
        follow_progress(
            ctx.obj.bi_tasks_api,
            [r.task_id for r in response.resources],
            [f"Create resource {r.friendly_name}" for r in response.resources],
            transient=True,
        )

    for resource in response.resources:
        console.print(f"[bold green]:sparkles: Created {resource.friendly_name} with ID {resource.id}")


@bi_app.command("list")
def list_experiment_resources(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
):
    """List experiment resources."""
    exp = _get_experiment(ctx, experiment)

    if exp.expires_at < datetime.now(timezone.utc):
        err_console.print(f"[red]Experiment {exp.friendly_name} has expired :x:")
        raise typer.Exit(1)

    experiment_resources = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
        experiment_id=exp.id,
        project=exp.project_id,
    )

    table = Table(
        "ID",
        "Name",
        "Status",
        "Created At",
        "Expires At",
        "Public IPv4",
        "Private IPv4",
        title=f"Resources in experiment {exp.id}",
    )
    for r in experiment_resources.resources:
        table.add_row(
            r.id,
            r.friendly_name,
            r.status,
            r.created_at.astimezone().strftime("%Y-%m-%d %H:%M"),
            r.expires_at.astimezone().strftime("%Y-%m-%d %H:%M") if r.expires_at else "-",
            r.public_ipv4,
            r.private_ipv4,
        )

    console.print(table)


def _get_resource(ctx: BIContext, experiment: Experiment, resource_id: str) -> ComputeResource:
    """Get a resource ID."""
    if _is_type_id(resource_id, required_type="r"):
        try:
            return ctx.obj.get_bi_resources_api_by_experiment(experiment).get_experiment_resource(
                experiment_id=experiment.id, project=experiment.project_id, resource_id=resource_id
            )
        except NotFoundException as err:
            err_console.print(f"[red]Resource {resource_id} not found in experiment {experiment.friendly_name} :x:")
            raise typer.Exit(code=1) from err

    resources = ctx.obj.get_bi_resources_api_by_experiment(experiment).get_experiment_resources(
        experiment_id=experiment.id, project=experiment.project_id, limit=100
    )

    while resources:
        for resource in resources.resources:
            if resource.friendly_name == resource_id:
                return resource

        marker = get_next_page_marker(resources.resource_links)
        resources = (
            ctx.obj.get_bi_resources_api_by_experiment(experiment).get_experiment_resources(
                experiment_id=experiment.id, project=experiment.project_id, marker=marker
            )
            if marker
            else None
        )

    err_console.print(f"[red]Resource {resource_id} not found in experiment {experiment.friendly_name} :x:")
    raise typer.Exit(code=1)


def _get_resource_id(ctx: BIContext, experiment: Experiment, resource_id: str) -> str:
    """Get a resource ID."""
    if _is_type_id(resource_id, required_type="r"):
        return resource_id

    resources = ctx.obj.get_bi_resources_api_by_experiment(experiment).get_experiment_resources(
        experiment_id=experiment.id, project=experiment.project_id, limit=100
    )

    while resources:
        for resource in resources.resources:
            if resource.friendly_name == resource_id:
                return resource.id

        marker = get_next_page_marker(resources.resource_links)
        resources = (
            ctx.obj.get_bi_resources_api_by_experiment(experiment).get_experiment_resources(
                experiment_id=experiment.id, project=experiment.project_id, marker=marker
            )
            if marker
            else None
        )

    err_console.print(f"[red]Resource {resource_id} not found in experiment {experiment.friendly_name} :x:")
    raise typer.Exit(code=1)


class PrintResourceDetail(Enum):
    """Proxy Options."""

    all = "all"
    resource_id = "resource_id"
    resource_name = "resource_name"
    exp_id = "exp_id"
    exp_name = "exp_name"
    proj_id = "proj_id"
    proj_name = "proj_name"
    status = "status"
    created = "created"
    expires = "expires"
    public_ipv4 = "public_ipv4"
    private_ipv4 = "private_ipv4"


class PrintResourceFormat(Enum):
    """Proxy Options."""

    text = "text"
    json = "json"
    yaml = "yaml"


@bi_app.command("print")
def print_resources(  # noqa: C901, PLR0912, PLR0915
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[
        Optional[str],
        typer.Argument(),
    ] = None,
    all_resources: Annotated[
        bool, typer.Option("--all", help="Destroy all resources in the experiment.", confirmation_prompt=True)
    ] = False,
    detail: Annotated[PrintResourceDetail, typer.Option()] = "all",  # pyright: ignore [reportArgumentType]  weird but this works
    format: Annotated[PrintResourceFormat, typer.Option()] = "text",  # pyright: ignore [reportArgumentType]  weird but this works
):
    """Print resource details."""

    if bool(resource) == bool(all_resources):
        err_console.print("[red]Please specify either --all or a resource name.")
        raise typer.Exit(code=1)

    exp = _get_experiment(ctx, experiment)

    def to_resource_dict(r: ComputeResource, add_id: bool) -> dict[str, str | None]:  # noqa: C901, PLR0912
        def _get_project_name():
            return "todo"  # TODO

        if detail == PrintResourceDetail.all:
            rd = {
                "resource_id": r.id,
                "resource_name": r.friendly_name,
                "status": r.status.name,
                "created": r.created_at.astimezone().isoformat() if r.created_at else None,
                "expires": r.expires_at.astimezone().isoformat() if r.expires_at else None,
                "public_ipv4": r.public_ipv4,
                "private_ipv4": r.private_ipv4,
                "exp_id": exp.id,
                "exp_name": exp.friendly_name,
                "project_id": exp.project_id,
                "project_name": _get_project_name(),
            }
        elif detail == PrintResourceDetail.resource_id:
            rd = {"resource_id": r.id}
        elif detail == PrintResourceDetail.resource_name:
            rd = {"resource_name": r.friendly_name}
        elif detail == PrintResourceDetail.status:
            rd = {"status": r.status.name}
        elif detail == PrintResourceDetail.created:
            rd = {"created": r.created_at.astimezone().isoformat() if r.created_at else None}
        elif detail == PrintResourceDetail.expires:
            rd = {"expires": r.expires_at.astimezone().isoformat() if r.expires_at else None}
        elif detail == PrintResourceDetail.public_ipv4:
            rd = {"public_ipv4": r.public_ipv4}
        elif detail == PrintResourceDetail.private_ipv4:
            rd = {"private_ipv4": r.private_ipv4}
        elif detail == PrintResourceDetail.proj_id:
            rd = {"project_id": exp.project_id}
        elif detail == PrintResourceDetail.proj_name:
            rd = {"proj_name": _get_project_name()}
        elif detail == PrintResourceDetail.exp_id:
            rd = {"exp_id": exp.id}
        elif detail == PrintResourceDetail.exp_name:
            rd = {"exp_name": exp.friendly_name}
        else:
            raise ValueError(f"Unimplemented detail requested: {detail}")
        if add_id:
            rd["resource_id"] = r.id
            rd["resource_name"] = r.friendly_name
            rd["exp_id"] = exp.id
            rd["exp_name"] = exp.friendly_name
        return rd  # pyright: ignore [reportReturnType] TODO

    if resource:
        r = _get_resource(ctx, exp, resource)
        res = to_resource_dict(r, add_id=False)
    else:
        resources = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
            exp.project_id, exp.id, limit=100
        )

        res = []
        while resources:
            for r in resources.resources:
                res.append(to_resource_dict(r, add_id=True))

            marker = get_next_page_marker(resources.resource_links)
            resources = (
                ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
                    experiment_id=exp.id, project=exp.project_id, marker=marker
                )
                if marker
                else None
            )

    if format == PrintResourceFormat.text:
        if resource:
            if len(res) > 1:
                for k, v in res.items():  # pyright: ignore [reportAttributeAccessIssue]
                    typer.echo(f"{k}={v}")
            else:
                for v in res.values():  # pyright: ignore [reportAttributeAccessIssue]
                    typer.echo(str(v))
        else:
            for r in res:
                for k, v in r.items():  # pyright: ignore [reportAttributeAccessIssue]
                    typer.echo(f"{k}={v}")
                typer.echo("\n")
    elif format == PrintResourceFormat.json:
        console.print(json.dumps(res))
    elif format == PrintResourceFormat.yaml:
        console.print(yaml.safe_dump(res))
    else:
        raise ValueError(f"Unimplemented format requested: {format}")


@bi_app.command("reset")
def reset_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[
        str,
        typer.Argument(
            ...,
        ),
    ],
):
    """Hard reset a resource."""
    exp = _get_experiment(ctx, experiment)
    res = _get_resource(ctx, exp, resource)

    task = ctx.obj.get_bi_resources_api_by_experiment(exp).perform_experiment_resource_action(
        exp.project_id,
        exp.id,
        res.id,
        ResourceAction(RebootAction(reboot=VmlabBackendSchemasBiResourceRebootActionAction(type="HARD"))),
    )

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    follow_one_progress(
        ctx.obj.bi_tasks_api,
        task.id,
        f"Resetting {resource}",
    )


def _destroy_one_resource(ctx: BIContext, exp: Experiment, res: ComputeResource):
    try:
        task = ctx.obj.get_bi_resources_api_by_experiment(exp).delete_experiment_resource(
            exp.project_id, exp.id, res.id
        )
    except NotFoundException as e:
        err_console.print(f"[red]{res.id} was not found :x:")
        raise typer.Exit(code=1) from e

    SSHKnownHostsManager().cleanup(res.id)

    follow_one_progress(
        ctx.obj.bi_tasks_api,
        task.id,
        f"Deleting {res.friendly_name}",
    )


def _destroy_all_experiment_resources(ctx: BIContext, exp: Experiment, skip_confirmation: bool = False):
    resources = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
        exp.project_id, exp.id, limit=100
    )

    delete_task_names_and_task_ids: list[tuple[str, str]] = []

    while resources:
        for res in resources.resources:
            if res.status in (ResourceState.DELETING, ResourceState.DELETED, ResourceState.FAILED):
                continue

            if not skip_confirmation:
                ok = Prompt.ask(
                    f"Are you sure that you want to [red bold]delete[/] resource [italic]{res.friendly_name}[/]?",
                    choices=["y", "n", "all"],
                )

                if ok == "n":
                    continue
                if ok == "all":
                    skip_confirmation = True

            delete_task = ctx.obj.get_bi_resources_api_by_experiment(exp).delete_experiment_resource(
                exp.project_id, exp.id, res.id
            )
            delete_task_names_and_task_ids.append((f"Deleting {res.friendly_name}", delete_task.id))

        marker = get_next_page_marker(resources.resource_links)
        resources = (
            ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
                experiment_id=exp.id, project=exp.project_id, marker=marker
            )
            if marker
            else None
        )

    follow_progress(
        ctx.obj.bi_tasks_api,
        [task_id for _, task_id in delete_task_names_and_task_ids],
        [task_name for task_name, _ in delete_task_names_and_task_ids],
    )


@bi_app.command("destroy")
def destroy_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[Optional[str], typer.Argument()] = None,
    all_resources: Annotated[
        bool, typer.Option("--all", help="Destroy all resources in the experiment.", confirmation_prompt=True)
    ] = False,
    force: Annotated[bool, typer.Option(help="Do not ask for confirmation.")] = False,
):
    """Destroy a resource."""

    if bool(resource) == bool(all_resources):
        err_console.print("[red]Please specify either --all or a resource name.")
        raise typer.Exit(code=1)

    exp = _get_experiment(ctx, experiment)

    if resource:
        res = _get_resource(ctx, exp, resource)
        _destroy_one_resource(ctx, exp, res)
        return

    _destroy_all_experiment_resources(ctx, exp, skip_confirmation=force)


@bi_app.command("renew")
def renew_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resources: Annotated[list[str], typer.Argument()],
    duration: Annotated[
        int,
        typer.Option(min=1, max=90 * 24, help="New lifetime of the resource in hours."),
    ],
):
    """Extend the expiration date of a resource."""
    exp = _get_experiment(ctx, experiment)
    res_ids = [_get_resource(ctx, exp, r).id for r in resources]

    new_expire_at = datetime.now(timezone.utc).replace(second=0, microsecond=0) + timedelta(hours=duration)

    # if experiment expiration is shorter than requested duration, extend the experiment
    if new_expire_at > exp.expires_at:
        err_console.print("Renewing Experiment...")
        ctx.obj.api.experiments_api.update_experiment(exp.id, ExperimentUpdate(expires_at=new_expire_at))

    if not len(res_ids):  # renew all resources if none are specified
        res_ids = [
            r.id
            for r in ctx.obj.get_bi_resources_api_by_experiment(exp)
            .get_experiment_resources(exp.project_id, exp.id, limit=100)
            .resources
        ]

    has_errors = False
    task_ids = []
    for res_id in res_ids:
        try:
            # err_console.print("Renewing Resource...")
            task = ctx.obj.get_bi_resources_api_by_experiment(exp).update_resource(
                exp.project_id, exp.id, res_id, UpdateResourcesRequest(expires_at=new_expire_at)
            )
        except NotFoundException:  # noqa: PERF203
            err_console.print(f"[red]{res_id} was not found :x:")
            has_errors = True
        else:
            task_ids.append(task.id)
            # err_console.print(
            #     f"[green]:heavy_check_mark: Done.[/green] Resource now expires at "
            #     f"{updated_resource.expires_at.astimezone().strftime('%Y-%m-%d %H:%M')}."
            # )

    follow_progress(
        ctx.obj.bi_tasks_api,
        task_ids,
        [f"Renewing Resource {exp.friendly_name} {resource}" for resource in resources],
    )

    if has_errors:
        raise typer.Exit(1)


class ProxyOption(Enum):
    """Proxy Options."""

    off = "off"
    on = "on"
    auto = "auto"


class ShowOption(Enum):
    """Show options."""

    nothing = "nothing"
    json = "json"
    command = "command"
    ansible = "ansible"
    ssh_config = "ssh_config"
    server_pubkey_openssh = "server_pubkey_openssh"
    proxy_pubkey_openssh = "proxy_pubkey_openssh"


DEFAULT_SSH_PORT = 22


def _can_connect(host, port) -> bool:
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.settimeout(0.1)
            return s.connect_ex((host, port)) == 0
    except:  # noqa: E722
        # Any exception means we can't connect.
        # Examples:
        #      gaierror: [Errno -5] No address associated with hostname
        return False


def _get_best_ssh_login(compute_resource: ComputeResource, with_proxy: bool):
    """Get the best SSH login for a resource."""
    if not compute_resource.ssh_logins:
        return None, False

    best_ssh_login = None
    best_ssh_login_can_connect = False
    best_score = -1

    # Prefer the first one that has a public key
    for ssh_login in compute_resource.ssh_logins:
        score = 0

        # User wants to use a proxy
        if with_proxy and ssh_login.jump_proxy:
            score += 1000

        # Prefer connecting to a public IP
        if compute_resource.public_ipv4 and compute_resource.public_ipv4 in ssh_login.host:
            score += 100

        # Prefer connecting without needing to use the proxy
        if can_connect := _can_connect(ssh_login.host, ssh_login.port):
            score += 10

        if score > best_score:
            best_ssh_login = ssh_login
            best_ssh_login_can_connect = can_connect
            best_score = score

    # Otherwise, just return the first one
    return best_ssh_login, best_ssh_login_can_connect


@bi_app.command("ssh")
def ssh_to_resource(  # noqa: C901, PLR0912, PLR0915
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
    no_exec: Annotated[
        bool,
        typer.Option(help="Do not call SSH to connect to the resource. Implied if --show is ansible or ssh_config."),
    ] = False,
    show: Annotated[
        ShowOption,
        typer.Option(
            help="What to show: Nothing, SSH Command, Ansible inventory line, "
            "lines for ~/.ssh/config, or the server public key"
        ),
    ] = "command",  # pyright: ignore [reportArgumentType]  weird but this works
    proxy: Annotated[ProxyOption, typer.Option(help="Use an SSH jump proxy.", case_sensitive=False)] = "auto",  # pyright: ignore [reportArgumentType]  weird but this works
    verbose: Annotated[int, typer.Option("-v", "--verbose", help="Verbose output.", count=True)] = 0,
):
    """Log in to a resource using SSH. Also show the login command."""
    exp = _get_experiment(ctx, experiment)
    compute_resource = _get_resource(ctx, exp, resource)

    ssh_login, can_connect = _get_best_ssh_login(compute_resource, with_proxy=proxy == ProxyOption.on)
    if not ssh_login:
        err_console.print("The resource has no SSH logins")
        raise typer.Exit(code=1)

    destination = f"{ssh_login.username}@{ssh_login.host}"
    if ssh_login.port != DEFAULT_SSH_PORT:
        destination += f":{ssh_login.port}"

    if proxy == ProxyOption.on and not ssh_login.jump_proxy:
        err_console.print("[red]Requested to use the proxy, but the resource has no jump proxy available.")
        err_console.print("Reverting to direct connection.")

    if proxy == ProxyOption.auto:
        proxy = (
            ProxyOption.off
            if ((compute_resource.public_ipv4 and compute_resource.public_ipv4 in ssh_login.host) or can_connect)
            else ProxyOption.on
        )

    # console.print(f"proxy={proxy} ssh_login.jump_proxy={ssh_login.jump_proxy}")

    ssh_command_options = []
    if verbose:
        ssh_command_options.append(f"-{'v' * verbose}")

    ansible_ssh_common_args = ""
    ssh_config_jump_proxy = ""
    if proxy == ProxyOption.on and ssh_login.jump_proxy:
        port_dest = f":{ssh_login.jump_proxy.port}" if ssh_login.jump_proxy.port != DEFAULT_SSH_PORT else ""
        ssh_jump_proxy_destination = f"{ssh_login.jump_proxy.username}@{ssh_login.jump_proxy.host}{port_dest}"
        ssh_command_options.extend(["-J", ssh_jump_proxy_destination])
        ansible_ssh_common_args = f" ansible_ssh_common_args='-J {ssh_jump_proxy_destination}'"
        ssh_config_jump_proxy = f"JumpProxy {ssh_jump_proxy_destination}"

    command = ["ssh", *ssh_command_options, destination]

    command_line = " ".join(command)

    do_exec_ssh = (not no_exec) and show in [ShowOption.command, ShowOption.nothing]

    if ssh_login.host_keys and compute_resource.status in (
        ResourceState.IMAGING,
        ResourceState.BOOTING,
        ResourceState.INITIALIZING,
        ResourceState.STARTING,
        ResourceState.UP,
    ):
        SSHKnownHostsManager().cleanup_expired()
        SSHKnownHostsManager().add_login(compute_resource, ssh_login)

    if show == ShowOption.command:
        if do_exec_ssh:
            console.print(f"[bold blue]{command_line}")
        else:
            typer.echo(command_line)  # output without any markup

    elif show == ShowOption.ansible:
        # TODO if ComputeResource ever has a user id for this host, use that
        # output without any markup
        typer.echo(
            f"{ssh_login.host} "
            f"ansible_ssh_port={ssh_login.port} "
            f"ansible_ssh_host={ssh_login.host} "
            f"ansible_ssh_user={ssh_login.username} {ansible_ssh_common_args}"
        )
    elif show == ShowOption.json:
        obj = {
            "host": ssh_login.host,
            "port": ssh_login.port,
            "username": ssh_login.username,
        }
        if ssh_login.jump_proxy:
            obj["jump_proxy"] = {
                "host": ssh_login.jump_proxy.host,
                "port": ssh_login.jump_proxy.port,
                "username": ssh_login.jump_proxy.username,
            }
        typer.echo(json.dumps(obj, indent=3))

    elif show == ShowOption.ssh_config:
        # output without any markup
        typer.echo(f"""Host={ssh_login.host}
   HostName {ssh_login.host}
   Port {ssh_login.port}
   User {ssh_login.username}
   PreferredAuthentications publickey
   IdentitiesOnly yes
   {ssh_config_jump_proxy}""")
    elif show == ShowOption.server_pubkey_openssh:
        if ssh_login.host_keys:
            for host_key in ssh_login.host_keys:
                typer.echo(f"{host_key.strip()}")
        else:
            err_console.print("No hostkeys in SSH info")
    elif show == ShowOption.proxy_pubkey_openssh:
        if ssh_login.jump_proxy:
            if ssh_login.jump_proxy.host_keys:
                for host_key in ssh_login.jump_proxy.host_keys:
                    typer.echo(f"{host_key.strip()}")
            else:
                err_console.print("No hostkeys in proxy SSH info")
        else:
            err_console.print("No jump proxy")

    if do_exec_ssh:
        os.execvp(command[0], command)  # noqa: S606


@bi_app.command("console")
def console_to_resource(  # noqa: C901
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
    show_only: Annotated[
        bool,
        typer.Option(help="Do not try to open console URL, only show it."),
    ] = False,
):
    """Open a browser with a console to a resource. Also print the link that is opened."""
    exp = _get_experiment(ctx, experiment)
    compute_resource = _get_resource(ctx, exp, resource)

    if not compute_resource.console_url:
        err_console.print("The resource has no remote console")
        raise typer.Exit(code=1)

    login: ConsoleLoginDetails | None
    try:
        login: ConsoleLoginDetails | None = ctx.obj.get_bi_resources_api_by_experiment(
            exp
        ).experiment_resource_get_console_login_details(
            experiment_id=compute_resource.experiment_id,
            project=exp.project_id,
            resource_id=compute_resource.id,
        )
    except NotFoundException:
        login = None

    if login:
        console.print(f"Console user: {login.username.get_secret_value()}")
        console.print(f"Console pass: {login.password.get_secret_value()}")
    console.print(f"Console URL: [bold blue]{compute_resource.console_url}")

    if not show_only:
        webbrowser.open(compute_resource.console_url, new=1)


def _userdata_file_relative_to_request_file(v, info: ValidationInfo):
    if v:
        assert info.context
        assert "file_path" in info.context
        assert isinstance(info.context["file_path"], str)

        return Path(info.context["file_path"]).parent / v
    return None


class ExtendedComputeResourceCreate(ComputeResourceCreate):
    """Extended ComputeResourceCreate."""

    userdata_file: Annotated[
        Optional[FilePath],
        PlainValidator(_userdata_file_relative_to_request_file),
        PlainSerializer(lambda _x: None),  # leave empty when serializing, otherwise dumping will fail.
    ] = None

    @model_validator(mode="after")
    def _check_userdata(self) -> "ExtendedComputeResourceCreate":
        if self.userdata and self.userdata_file:
            raise ValueError("Both userdata and userdata_file are set. Only one can be used.")
        return self

    @field_validator("disk_image_id")
    def disk_image_id_validate_regular_expression(cls, value):
        """Disable validation."""
        return value

    @field_validator("flavor_id")
    def flavor_id_validate_regular_expression(cls, value):
        """Disable validation."""
        return value

    site_id: Annotated[
        Optional[str],
        Field(description="The site on which the resource must be provisioned"),
    ] = None


class Request(BaseModel):
    """Compute Resources Request."""

    resources: list[ComputeResourceCreate]
    links: Optional[list[BareMetalLinkCreate]] = None


class ExtendedRequest(BaseModel):
    """Extended Compute Resources Request."""

    resources: list[ExtendedComputeResourceCreate]
    links: Optional[list[BareMetalLinkCreate]] = None


def _process_extended_resource(ctx: BIContext, resource: ExtendedComputeResourceCreate) -> ComputeResourceCreate:
    disk_image_id = resource.disk_image_id
    if disk_image_id and not _is_type_id(disk_image_id, required_type="image"):
        disk_image_id = _find_disk_image_id(ctx, disk_image_id)

    flavor_id = resource.flavor_id
    if flavor_id and not _is_type_id(resource.flavor_id, required_type="flavor"):
        flavor_id = _find_flavor_id(ctx, flavor_id)

    userdata = resource.userdata
    if resource.userdata_file:
        assert not userdata
        with Path.open(resource.userdata_file) as f:
            userdata = f.read()

    if resource.site_id:
        if resource.site_id not in ctx.obj.api.config.bi:
            err_console.print(f"[red]Site {resource.site_id} is unknown.")
        if disk_image_id and not disk_image_id.startswith(f"image_{resource.site_id}_"):
            err_console.print(f"[red]Disk image {disk_image_id} does not belong to site {resource.site_id}.")
        if flavor_id and not flavor_id.startswith(f"flavor_{resource.site_id}_"):
            err_console.print(f"[red]Disk image {disk_image_id} does not belong to site {resource.site_id}.")

    return ComputeResourceCreate(
        friendly_name=resource.friendly_name,
        disk_image_id=disk_image_id,
        flavor_id=flavor_id,
        network_interfaces=resource.network_interfaces,
        userdata=userdata,
    )


def _read_and_process_request_file(ctx: BIContext, file: typer.FileText) -> Request:
    try:
        extended_request = ExtendedRequest.model_validate_json(file.read(), context={"file_path": file.name})
    except ValidationError as e:
        err_console.print("[red]Errors in request file:")
        for error in e.errors(include_url=False):
            err_console.print(f"[red]{'.'.join(str(loc) for loc in error['loc'])}:")
            err_console.print(f"[red]  {error['msg']}\n")
        raise typer.Exit(code=1) from e

    site_ids = {r.site_id for r in extended_request.resources if r.site_id}
    if len(site_ids) > 1:
        err_console.print("[red]Only single-site requests are supported for now.")
        raise typer.Exit(code=1)

    return Request(
        resources=[_process_extended_resource(ctx, er) for er in extended_request.resources],
        links=extended_request.links,
    )


@bi_app.command()
def create_from_file(
    ctx: BIContext,
    experiment: Annotated[str, typer.Option(..., help="Experiment name or ID.", prompt=True)],
    file: Annotated[typer.FileText, typer.Option(..., help="File with resource definitions.", prompt=True)],
    duration: Annotated[
        int, typer.Option(min=1, max=90 * 24, help="Lifetime of the resources in hours. Can be extended later on.")
    ] = 24,
    wait: Annotated[bool, typer.Option(help="Wait for resources to be created.")] = False,
):
    """Request compute resources."""
    request = _read_and_process_request_file(ctx, file)

    expires_at = datetime.now(timezone.utc).replace(second=0, microsecond=0) + timedelta(hours=duration)

    project_id = ctx.obj.require_project_id()

    if is_type_id(experiment, type="exp"):
        exp = ctx.obj.api.get_experiment_by_id(experiment_id=experiment)
    else:
        exp = ctx.obj.api.get_or_create_experiment(
            experiment_name=experiment,
            project_id=project_id,
            expires_at=expires_at,
        )

    # if experiment expiration is shorter than requested duration, extend the experiment
    if expires_at > exp.expires_at:
        ctx.obj.api.experiments_api.update_experiment(exp.id, ExperimentUpdate(expires_at=expires_at))

    ssh_authorized_keys = _get_registered_pubkeys(ctx, exp)
    if not ssh_authorized_keys:
        err_console.print("[red]You need to have at least one public SSH key registered. :x:")
        err_console.print("Use: [blue]slices pubkey register")
        raise typer.Exit(code=1)

    nrr = NewResourcesRequest(
        expires_at=expires_at,
        resources=request.resources,
        links=request.links,
        ssh_authorized_keys=ssh_authorized_keys,
        tickets=[
            Ticket(
                jwt=ctx.obj.api.get_id_token_for_audience(ctx.obj.api.config.bi[ctx.obj.site_id].audience, project_id)
            )
        ],
    )

    try:
        response = ctx.obj.get_bi_resources_api_by_experiment(exp).request_experiment_resources(
            project=project_id,
            experiment_id=exp.id,
            new_resources_request=nrr,
        )
    except ForbiddenException as err:
        err_console.print("[bold red]No permission to create resource[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None
    except ApiException as err:
        err_console.print("[bold red]Could not create resource[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None
    if not len(response.resources):
        err_console.print("[bold red]No resources were created[/] :x:")

    if wait:
        follow_progress(
            ctx.obj.bi_tasks_api,
            [r.task_id for r in response.resources],
            [f"Create resource {r.friendly_name}" for r in response.resources],
            transient=True,
        )

    for resource in response.resources:
        console.print(f"[bold green]:sparkles: Created {resource.friendly_name} with ID {resource.id}")
